import os.path
import numpy as np
import pandas as pd
from statsmodels.genmod.generalized_linear_model import GLM
from statsmodels.genmod import families

import statsmodels.stats.tests.test_influence
test_module = statsmodels.stats.tests.test_influence.__file__
cur_dir = cur_dir = os.path.abspath(os.path.dirname(test_module))

from sklearn.decomposition import PCA


np.random.seed(1234)
lambs,sens = [],[]
for spec_id in range(50):
    lam = 1*np.random.gamma(1.5)
    lambs.append(lam)
    sens.append( (np.random.random()*4 - 2))

    
np.random.seed(None)


NUM_ROWS = 100

nn = 2*np.random.random(size=[NUM_ROWS]) -1
XX = np.random.random([NUM_ROWS])*1.5+0.5


xdf = pd.DataFrame()
xdf['c'] = 1 + np.zeros_like(nn)
xdf['X'] = XX

datasets = []


NOISE_MODE = "PCA"
#NOISE_MODE = "REG"

for expt in range(20):
    ydf = pd.DataFrame()
    mudf = pd.DataFrame()
    #nn = 2*np.random.random(size=[150]) -1
    #XX = np.random.random([150])*1.5+0.5
    for spec_id in range(50):
        lam = lambs[spec_id]
        lamv = lam*(XX)
        sensi = sens[spec_id]
        ydf['obs_' + str(spec_id)] = np.random.poisson(np.exp(nn*sensi + lamv + 
                                                        (2*np.random.random(size=[NUM_ROWS])-1)*0.0 ))
        mudf['obs_' + str(spec_id)] = np.exp( lamv )
    xdf = pd.DataFrame()
    xdf['c'] = 1 + np.zeros_like(nn)
    xdf['X'] = XX


    datasets.append((ydf,mudf,xdf))


#print (mudf)


def compute_resids(dataset, devmode, family, transform):
    resds = []
    bpreds = []
    N = 50
    ydf = dataset[0]
    xdf = dataset[2]
    for spec_id in range(N):
        res = GLM( transform(ydf['obs_' + str(spec_id)]), xdf, family=family).fit(attach_wls=True, atol=1e-4)
        infl = res.get_influence(observed=True)
        summ_df = infl.summary_frame()
        if devmode == "YdA":
            resds.append( (ydf['obs_' + str(spec_id)] - res.predict(xdf))/res.predict(xdf)  )
        elif devmode == "Yd":
            resds.append( ydf['obs_' + str(spec_id)] - res.predict(xdf) )
        elif devmode == "pp":
            resds.append( (ydf['obs_' + str(spec_id)] - res.predict(xdf) )/np.sqrt(res.predict(xdf)) )
        elif devmode == "rd":
            resds.append( res.resid_deviance)
        elif devmode == "std":
            resds.append( summ_df['standard_resid'] )
        else:
            resds.append( 0*summ_df['standard_resid'] )

    return resds




def sglm(dataset, devmode, num_resids=5, family=families.Poisson(), transform=lambda x: x, inv_transform=lambda x: x):

    resds = compute_resids(dataset, devmode, family, transform)
    #print (np.array(resds).shape)
    NN = np.array(resds[25:25+num_resids]).transpose()
    #print (NN.shape)
    N = 50
    bpreds = []

    xdf = dataset[2]
    x1df = xdf.copy()
    x2df = xdf.copy()





    if devmode is None or NN.shape[0] == 0:
        noise_estim = 0*nn 
    else:
        pca = PCA(n_components=1)
        NN = NN/(1e-6+np.std(NN,axis=0))
        noise_estim = pca.fit_transform(NN)
        noise_res = GLM(noise_estim, xdf, family=families.Gaussian()).fit( atol=1e-4)

    x1df['n'] = noise_estim
    x2df['n'] = 0

    ydf = dataset[0]

    preds = []

    if num_resids > 0:
        NN = np.concatenate([NN, np.ones([NN.shape[0],1]), xdf.values ], axis=1)
        alpha = np.array([2 for _ in range(25,25+num_resids)] + [10])

    for spec_id in range(25):
        if NOISE_MODE == "REG" and num_resids > 0:

            noise_res = GLM(resds[spec_id], NN, family=families.Gaussian()).fit(attach_wls=True, atol=1e-4) #.fit_regularized(alpha=0*alpha)
            noise_X = GLM(resds[spec_id], xdf, family=families.Gaussian()).fit(attach_wls=True, atol=1e-4)
            x1df['n'] = noise_res.predict(NN) - noise_X.predict(xdf)

        res = GLM(transform(ydf['obs_' + str(spec_id)]), x1df, family=family).fit(attach_wls=True, atol=1e-4)
        preds.append(( inv_transform(res.predict(x2df))))

    return np.array(preds).transpose()


def vs3qs(dataset, num_resids=5):
    family = families.Gaussian()
    transform = lambda x: np.sqrt(x)
    inv_transform = lambda x: x**2
    return sglm(dataset, 'Yd', num_resids, family=family, transform=transform, inv_transform=inv_transform)


def vs3qs(dataset, num_resids=5):
    family = families.Gaussian()
    transform = lambda x: np.log(x+1)
    inv_transform = lambda x: np.exp(x) - 1
    return sglm(dataset, 'Yd', num_resids, family=family, transform=transform, inv_transform=inv_transform)






plot_bias_dict = {}
plot_var_dict = {}
plot_mse_dict = {}

for mmd in ['rd', 'YdA', 'Yd', 'std', 'pp']:
    out_dict = {}
    out_dict_pred = {}
    num_resid_list = [0,1,3,7,10,15,20]
    for num_resids in num_resid_list:
        out_dict[num_resids] = []
        out_dict_pred[num_resids] = []
        for sim in range(20):
            preds = sglm(datasets[sim], mmd, num_resids)
            mus = datasets[sim][1][['obs_' + str(spec_id) for spec_id in range(25)]]
            XXs = datasets[sim][2]['X']
            #print (preds.shape, mus.shape)
            err = (np.log(preds) - np.array(np.log(mus)))/(1.0 + 0*mus) #Hack to make it a pandas array
            #print (err)
            out_dict[num_resids].append(err.values)
            out_dict_pred[num_resids].append(np.log(preds))
    print [mmd] + [ (k,np.mean(out_dict[k])) for k in num_resid_list]
    print [mmd] + [ (k,np.mean(np.std(np.array(out_dict_pred[k]), axis=0))) for k in num_resid_list]
    print [mmd] + [ (k,np.std(np.mean(np.array(out_dict_pred[k]), axis=(1,2)))) for k in num_resid_list]
    plot_bias_dict[mmd] = [ np.mean(out_dict[k]) for k in num_resid_list]
    plot_var_dict[mmd] = [ np.std(np.mean(np.array(out_dict_pred[k]), axis=(1,2))) for k in num_resid_list]

    plot_mse_dict[mmd] = [ np.mean(np.array(out_dict[k])**2) for k in num_resid_list]

out_dict = {}
out_dict_pred = {}
num_resid_list = [0,1,3,7,10,15,20]
for num_resids in num_resid_list:
    out_dict[num_resids] = []
    out_dict_pred[num_resids] = []
    for sim in range(20):
        preds = vs3qs(datasets[sim], num_resids)
        mus = datasets[sim][1][['obs_' + str(spec_id) for spec_id in range(25)]]
        err = np.log(np.clip(preds, 0, None) + 1e-3) - np.array(np.log(mus + 1e-3))
        #err = (preds) - np.array((mus))
        out_dict[num_resids].append(err)
        out_dict_pred[num_resids].append(np.log(np.clip(preds, 0, None) + 1e-3))

print [ (k,np.mean(out_dict[k])) for k in num_resid_list]
print [ (k,np.mean(np.std(np.array(out_dict_pred[k]), axis=0))) for k in num_resid_list]
print [ (k,np.std(np.mean(np.array(out_dict[k]), axis=(1,2)))) for k in num_resid_list]

plot_bias_dict['3qs'] = [ np.mean(out_dict[k]) for k in num_resid_list]
plot_var_dict['3qs'] = [ np.std(np.mean(np.array(out_dict_pred[k]), axis=(1,2))) for k in num_resid_list]

plot_mse_dict['3qs'] = [ np.mean(np.array(out_dict[k])**2) for k in num_resid_list]

from sklearn.metrics import roc_curve, auc
import matplotlib.pyplot as plt
from itertools import cycle

plt.figure()
lw = 2

rndict = {'YdA': '$R_{SGLM}$', '3qs': '$R_{3qs}$', 'rd' : '$R_{Deviance}$', 'std': '$R_{Student}$', 'Yd': '$R_{Raw}$'}


for kk in rndict.keys():
    plt.plot( num_resid_list , np.array(plot_bias_dict[kk]) ,'-', markersize=3, label=rndict[kk])
plt.xlabel('n', fontsize=20)
plt.ylabel('Bias', fontsize=20)
plt.yticks(fontsize=14)
plt.xticks(fontsize=16)
    
#plt.title('%s' % "Bias", fontsize=10)
plt.legend(fontsize=16)
plt.savefig('poisson_bias_sim_appx.pdf', bbox_inches='tight', pad_inches=0.05)
plt.close()



for kk in rndict.keys():
    plt.plot( num_resid_list , np.array(plot_mse_dict[kk]) ,'-', markersize=3, label=rndict[kk])
plt.xlabel('n', fontsize=20)
plt.ylabel('MSE', fontsize=20)
plt.xticks(fontsize=18)
plt.yticks(fontsize=14)
    
#plt.title('%s' % "Bias", fontsize=10)
plt.legend(fontsize=18)
plt.savefig('poisson_mse_sim_appx.pdf', bbox_inches='tight', pad_inches=0.05)
plt.close()



for kk in rndict.keys():
    if kk == '3qs':
        continue
    plt.plot( num_resid_list , np.array(plot_mse_dict[kk]) ,'-', markersize=3, label=rndict[kk])
plt.xlabel('n', fontsize=20)
plt.ylabel('MSE', fontsize=20)
plt.xticks(fontsize=18)
plt.yticks(fontsize=14)
    
#plt.title('%s' % "Bias", fontsize=10)
plt.legend(fontsize=18)
plt.savefig('poisson_mse_sim_appx1.pdf', bbox_inches='tight', pad_inches=0.05)
plt.close()
